/* $Id$ 
 *
 * Copyright   : (c) 1998 by Glenn Trigg.  All Rights Reserved
 * Project     : bib - Motif bible search tool.
 * File        : bib_misc
 *
 * Author      : Glenn Trigg
 * Created     :  1 Apr 1998 
 *
 * Description : 
 */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<rxposix.h>

#include	"bib_books.h"

static void print_summary(void);

#ifndef CHECK_INDEX
static char *fullnames[] = {
    "Genesis", "Exodus", "Leviticus", "Numbers", "Deuteronomy",
    "Joshua", "Judges", "Ruth", "I Samuel", "II Samuel", "I Kings",
    "II Kings", "I Chronicles", "II Chronicles", "Ezra", "Nehemiah",
    "Esther", "Job", "Psalms", "Proverbs", "Ecclesiastes", "Songs of Sol.",
    "Isaiah", "Jeremiah", "Lamentations", "Ezekiel", "Daniel", "Hosea",
    "Joel", "Amos", "Obadiah", "Jonah", "Micah", "Nahum", "Habakkuk",
    "Zephaniah", "Haggai", "Zechariah", "Malachi", "Matthew", "Mark",
    "Luke", "John", "The Acts", "Romans", "I Corinthians", "II Corinthians",
    "Galations", "Ephesians", "Phillipians", "Colossians", "I Thessalonians",
    "II Thessalonians", "I Timothy", "II Timothy", "Titus", "Philemon",
    "Hebrews", "James", "I Peter", "II Peter", "I John", "II John",
    "III John", "Jude", "Revelation"
};
#endif

/* Find the absolute verse number given the book (Gen = 1), chapter & verse. */
/* (The books[].abbrev will probably come in useful for going from something
   like 2Cor to the actual book number. Return -1 for errors. */
long
find_verse(int book, int chapter, int verse)
{
    int		i;
    int		vcount = 0;	/* verse count */

    /* Need to find the total number of verses preceeding this reference. */

    /* First tally up all complete books before the given one. */
    if((book < 1) || (book > 66)) {
	return -1L;		/* Error condition. */
    }
    for(i = 0; i < book - 1; i++) {
	vcount += books[i].total_verses;
    }

    /* Then tally up the chapters before the given one. */
    if((chapter < 1) || (chapter > books[book - 1].num_chapters)) {
	return -1L;		/* Error condition. */
    }
    for(i = 0; i < chapter - 1; i++) {
	vcount += books[book - 1].verse_count[i];
    }

    /* Then the verses. */
    if((verse < 1) || (verse > books[book - 1].verse_count[chapter - 1])) {
	return -1L;
    }
    vcount += verse - 1;

    return verse_index[vcount];		/* This is then used with the big array */
				/* of file positions for each verse to */
				/* position the file pointer at the required */
				/* verse. */
}

/* Take a book abbreviation and return the book number. Eg:
   Gen -> 1, Exo -> 2, Rev -> 66, etc. */
int
book_num(char *abbrev)
{
    int		i;

    for(i = 0; i < 66; i++) {
	if(strcasecmp(abbrev, books[i].abbrev) == 0)
	    return(i + 1);
    }

    return -1;			/* -1 indicates failure. */
}

#ifndef CHECK_INDEX
void
init_books(char *bible_name)
{
    int		ret, book_num = 0, chap_num = 0, i, j, vtot;
    int		vix_count = 0;
    char	curr_book[8], curr_chap[8], book[8], chap[4], vrse[4];
    char	buf[BUFSIZ];
    FILE	*fp;
    regex_t	re;
    regmatch_t	rm[4];

    ret = regcomp(&re, "^([[:digit:]]?[[:alpha:]]+)([[:digit:]]+):([[:digit:]]+) ", REG_EXTENDED);

    if( ret != 0 ) {
	fprintf(stderr, "Couldn't compile regexp.\n");
	return;
    }

    curr_book[0] = '\0';	/* Initialise */
    curr_chap[0] = '\0';

    if( (fp = fopen(bible_name, "r")) == NULL ) {
	fprintf(stderr, "Couldn't open bible file (%s).\n", bible_name);
	return;
    }
    rewind(fp);			/* Just to be paranoid! :-) */

    verse_index[vix_count++] = ftell(fp);

    while(fgets(buf, BUFSIZ-1, fp) != NULL) {
	verse_index[vix_count++] = ftell(fp);
ret = regexec(&re, buf, 4, rm, 0);
	if(ret != 0) {
	    fprintf(stderr, "Error matching line: %s\n", buf);
	    continue;
	}

	strncpy(book, &buf[rm[1].rm_so], rm[1].rm_eo - rm[1].rm_so);
	book[rm[1].rm_eo - rm[1].rm_so] = '\0';

	strncpy(chap, &buf[rm[2].rm_so], rm[2].rm_eo - rm[2].rm_so);
	chap[rm[2].rm_eo - rm[2].rm_so] = '\0';

	if( strcmp(curr_book, book) != 0 ) {
	    if(book_num != 0) {
		/* Finish off the last book. */
		books[book_num - 1].num_chapters = atoi(curr_chap);
		fprintf(stderr, "<%s>", vrse);
		books[book_num - 1].verse_count[atoi(curr_chap) - 1] = atoi(vrse);
		chap_num = 0;	/* reset chapter number */
	    }
	    book_num++;
	    strcpy(curr_book, book);
	    strcpy(books[book_num - 1].abbrev, book);
	    fprintf(stderr, "\nDoing book: %s", curr_book);
	}
	if( (strcmp(curr_chap, chap) != 0) || (chap_num == 0) ) {
	    if(chap_num != 0) {
		/* Finish off last chapter. */
		fprintf(stderr, "<%s>", vrse);
		books[book_num - 1].verse_count[atoi(curr_chap) - 1] = atoi(vrse);
	    }
	    chap_num++;
	    strcpy(curr_chap, chap);
	    fprintf(stderr, " %s", curr_chap);
	}

	strncpy(vrse, &buf[rm[3].rm_so], rm[3].rm_eo - rm[3].rm_so);
	vrse[rm[3].rm_eo - rm[3].rm_so] = '\0';
    }

    fclose(fp);

    /* Finish off last verse. */
    books[book_num - 1].num_chapters = atoi(curr_chap);
    fprintf(stderr, "<%s>", vrse);
    books[book_num - 1].verse_count[atoi(curr_chap) - 1] = atoi(vrse);
    fprintf(stderr, "\n");

    /* Print summary & calc total verses. */
    for(i = 0; i < 66; i++) {
	strcpy(books[i].fullname, fullnames[i]);
	vtot = 0;
	for(j = 0; j < books[i].num_chapters; j++) {
	    vtot += books[i].verse_count[j];
	}
	books[i].total_verses = vtot;
    }

    print_summary();
}

static void
write_index(char *index_name)
{
    FILE	*fp;

    if( (fp = fopen(index_name, "w")) == NULL ) {
	fprintf(stderr, "Can't write to index file (%s).\n", index_name);
	return;
    }

    if( fwrite(verse_index, sizeof(verse_index), 1, fp) != 1 ) {
	fprintf(stderr, "Error writing index array, incomplete index file!\n");
	return;
    }
    if( fwrite(books, sizeof(books), 1, fp) != 1 ) {
	fprintf(stderr, "Error writing book info, incomplete index file!\n");
	return;
    }

    fclose(fp);
}
#endif

static void
read_index(char *index_name)
{
    FILE	*fp;

    if( (fp = fopen(index_name, "r")) == NULL ) {
	fprintf(stderr, "Can't read from index file (%s).\n", index_name);
	return;
    }

    if( fread(verse_index, sizeof(verse_index), 1, fp) != 1 ) {
	fprintf(stderr, "Error reading index array, incomplete index file!\n");
	return;
    }
    if( fread(books, sizeof(books), 1, fp) != 1 ) {
	fprintf(stderr, "Error reading book info, incomplete index file!\n");
	return;
    }

    fclose(fp);
}

static void
print_summary()
{
    int		i, j, vtot = 0;

    /* Print summary & calc total verses. */
    printf("\n\nSummary:\n");

    for(i = 0; i < 66; i++) {
	printf("Book %d: %s [%s] (%d chaps)", i,
	       books[i].fullname, books[i].abbrev,
	       books[i].num_chapters);
	for(j = 0; j < books[i].num_chapters; j++) {
	    printf(" %d", books[i].verse_count[j]);
	}
	vtot += books[i].total_verses;
	printf(" [%d total %d run]\n", books[i].total_verses, vtot);
    }
    printf("Grand verse total: %d\n", vtot);
}

int
main(int argc, char **argv)
{
#ifndef CHECK_INDEX
    init_books("bib.data");
    write_index("bib.data.idx");
#else
    char	abbrev[20], verse[BUFSIZ];
    int		ch, vs;
    long	offset;
    FILE	*fp;

    read_index("bib.data.idx");
    print_summary();

    if( (fp = fopen("bib.data", "r")) == NULL ) {
	fprintf(stderr, "Warning: couldn't open bib.data\n");
    }

    while( ! feof(stdin) ) {
	printf("Enter book abbreviation, chapter & verse: ");
	if( scanf("%s %d %d", abbrev, &ch, &vs) < 3 ) {
	    printf("\n");
	    break;
	}

	printf("Verse seek value: %ld\n", offset = find_verse(book_num(abbrev),
							 ch, vs));
	if(fp != NULL) {
	    if( offset == -1L ) {
		puts("Invalid reference.\n");
	    }
	    else {
		fseek(fp, offset, 0);
		fgets(verse, BUFSIZ-1, fp);
		puts(verse);
	    }
	}
    }
    if(fp != NULL)
	fclose(fp);
#endif

    return 0;
}
